import os
from typing import List, Optional
from pydantic import BaseModel
from enum import Enum

import yaml

# data model

DIR, _ = os.path.split(os.path.abspath(__file__))
with open(os.path.join(DIR, 'enum_translation.yml')) as f:
    trans_data = yaml.safe_load(f)

class Location(BaseModel):
    lat: float
    lon: float

class Rating(BaseModel):
    avg_rating: float
    avg_on_time: float
    avg_service: float
    avg_accurate_profile_description: float


class BaseProfile(BaseModel):
    wallet: Optional[str]
    uid: Optional[str]
    name: str
    referrer: str = None
    gender: str = None
    register_timestamp: int = 0


class MultiLangEnum(str, Enum):
    @classmethod
    def translates(cls):
        return trans_data[cls.__name__]

class Gender(MultiLangEnum):
    male = 'male'
    female = 'female'

class EyeColor(MultiLangEnum):
    blue = 'blue'
    black = 'black'
    brown = 'brown'
    green = 'green'
    other = 'other'

class HairColor(MultiLangEnum):
    blue = 'blue'
    black = 'black'
    blond = 'blond'
    red = 'red'
    other = 'other'

class Ethnicity(MultiLangEnum):
    asian = 'asian'
    australian = 'australian'
    caucasian = 'caucasian'

class Build(MultiLangEnum):
    athletic = 'athletic'
    chunky = 'chunky'
    fit = 'fit'
    slender = 'slender'
    skinny = 'skinny'


class Bust(MultiLangEnum):
    a = 'a'
    b = 'b'
    c = 'c'
    d = 'd'
    e = 'e'
    f = 'f'
    g_plus = 'g_plus'


class DressSize(MultiLangEnum):
    small_minus = 'small_minus'
    small = 'small'
    medium = 'medium'
    large = 'large'
    large_plus = 'large_plus'

class SpeakingLanguage (MultiLangEnum):
    english = 'english'
    mandarin = 'mandarin'
    japanese = 'japanese'
    korean = 'korean'
    cantonese = 'cantonese'

class PaymentMethod (MultiLangEnum):
    usdt = 'usdt'

class TimeSlotStatus(MultiLangEnum):
    available = 'available'
    booked = 'booked'
    locked = 'locked'


class BookingStatus(MultiLangEnum):
    attempt = 'attempt'
    archived = 'archived'
    confirmed = 'confirmed'
    cancel_attempt = 'cancel_attempt'
    canceled = 'canceled'
    fulfilled = 'fulfilled'
    deleted = 'deleted'  # need this? or simply delete it. as _id will conflict


class ProviderProfile(BaseProfile):
    address: str
    postcode: Optional[int]
    city: Optional[str]
    country: Optional[str]
    age: int = 27
    location: Location
    contact_detail: str = None
    rate_aud: int = 150
    hair_color: Optional[HairColor]
    build: Optional[Build]
    ethnicity: Optional[Ethnicity]
    eye_color: Optional[EyeColor]
    bio: Optional[str]
    photos: List[str] = []
    height: Optional[int]
    bust: Optional[Bust]
    avg_on_time: Optional[float]
    avg_service: Optional[float]
    avg_rating: Optional[float]
    avg_accurate_profile_description: Optional[float]
    dress_size: Optional[DressSize]
    speaking_language: List[SpeakingLanguage] = []
    payment: List[PaymentMethod] = []


class ConsumerProfile(BaseProfile):
    pass


class TimeSlot(ProviderProfile):
    slot_id: int  # the slot id, YYYYmmddXX
    slot_status: TimeSlotStatus = TimeSlotStatus.available


# details of a booking, which is shown to the provider and consumer
class BookingDetail(TimeSlot):
    total_fee_aud: int


class BookingHistory(BaseModel):  # state chagen history of a booking
    ationer: str
    timestamp: int
    additional_comment: Optional[str]


class Booking(BaseModel):  # booking to a timeslot
    consumer_uid: str
    provider_uid: str
    all_slots: List[int] = []
    status: BookingStatus
    consumer_comments: Optional[str]
    consumer_rating: Optional[float]
    provider_comments: Optional[str]
    provider_rating: Optional[float]
    last_update: int
    detail: BookingDetail
    total_fee_aud: int
    book_time: int  # epoch second of booked
    history: List[BookingHistory] = []


class Transaction(BaseModel):
    consumer_uid: str
    provider_uid: str
    booking: Booking
    timestamp: int
    total_fee_aud: int


# response model


class SingleProviderResponse(ProviderProfile):
    pass


class SingleConsumerResponse(ConsumerProfile):
    pass


class SingleTimeSlotResponse(TimeSlot):
    pass


class SingleBookingResponse(Booking):
    pass


class SingleTransactionResponse(Transaction):
    pass


class ProviderListResponse(BaseProfile):
    results: List[ProviderProfile] = []
    start: Optional[int]
    total_size: Optional[int]
    next_cursor: Optional[str]


class TimeSlotListResponse(BaseProfile):
    results: List[TimeSlot] = []
    start: Optional[int]
    total_size: Optional[int]
    next_cursor: Optional[str]


class BookingListResponse(BaseProfile):
    results: List[Booking] = []
    start: Optional[int]
    total_size: Optional[int]
    next_cursor: Optional[str]


class OrderListResponse(BaseProfile):
    results: List[Booking] = []
    start: Optional[int]
    total_size: Optional[int]
    next_cursor: Optional[str]

# general model


class HealthResponse(BaseModel):
    status: str

class NotificationsModel(BaseModel):
    notifications: List[str] = []


"""
Elastic Helper for Bdating.
"""
from fastapi import HTTPException
from elasticsearch import AsyncElasticsearch


class BdatingES(AsyncElasticsearch):
    def __init__(self, settings):
        super().__init__(settings.es_endpoint)
        self.settings = settings

    async def get(self, *args, **kw):
        kw.setdefault('index', self.settings.es_index)
        result = await super().get(*args, **kw)
        return es_get_result_to_dict(result)

    async def update(self, *args, **kw):
        kw.setdefault('index', self.settings.es_index)
        result = await super().update(*args, **kw)
        return result

    async def search(self, *args, **kw):
        kw.setdefault('index', self.settings.es_index)
        result = await super().search(*args, **kw)
        return es_search_result_to_dict(result)

    async def delete(self, *args, **kw):
        kw.setdefault('index', self.settings.es_index)
        result = await super().delete(*args, **kw)
        return result

    async def delete_by_query(self, *args, **kw):
        kw.setdefault('index', self.settings.es_index)
        result = await super().delete_by_query(*args, **kw)
        return result

    async def create(self, *args, **kw):
        kw.setdefault('index', self.settings.es_index)
        result = await super().create(*args, **kw)
        return result
    
    async def index(self, *args, **kw):
        kw.setdefault('index', self.settings.es_index)
        result = await super().index(*args, **kw)
        return result

def es_get_result_to_dict(es_result: dict, with_id: bool = False) -> dict:
    """
    A typical es result
    {
    "_index": "local-bdating",
    "_id": "a0abCd:provider",
    "_version": 1,
    "_seq_no": 0,
    "_primary_term": 1,
    "found": true,
    "_source": {
      "type": "provider",
      "uid":"23fdad",
      "address": "10809 Mercedes Loaf Apt. 552\nNancyborough, ID 01277",
      "location": [
        -38.1355732927354,
        145.21265948643753
      ],
      "name": "Julie Simmons",
      "bio": "Number next prepare possible generation among arm."
    }
  }
    """
    result = es_result.get('_source', {})
    if with_id:
        result['_id'] = es_result.get('_id')
    return result


def es_search_result_to_dict(es_result: dict, with_id: bool = False) -> dict:
    '''
    A typical es result
    {
  "took" : 6,
  "timed_out" : false,
  "_shards" : {
    "total" : 1,
    "successful" : 1,
    "skipped" : 0,
    "failed" : 0
  },
  "hits" : {
    "total" : {
      "value" : 10,
      "relation" : "eq"
    },
    "max_score" : 0.020639554,
    "hits" : [
      {
        "_index" : "local-bdating",
        "_id" : "a0abCd:slot",
        "_score" : 0.020639554,
        "_source" : {
          "type" : "slot",
          "uid" : "a0abCd",

          "address" : """Unit 9190 Box 6681


DPO AP 42608""",
          "location" : {
            "lat" : -37.977931083798815,
            "lon" : 145.07269918572806
          },
          "name" : "Zachary Powers",
          "bio" : "Type seat some agent.",
          "rate_aud" : 150,
          "slot_id" : 2022062623
        }
      },
      ...
    ]
  }
}

  }
    '''
    result = {
        'results': []
    }
    for r in es_result.get('hits', {})['hits']:
        record = r.get('_source')
        if with_id:
            record['_id'] = r.get('_id')
        result['results'].append(record)

    if es_result.get('hits', {}).get('total', {}).get('value') is not None:
        result['total_size'] = es_result.get(
            'hits', {}).get('total', {}).get('value')
    result['size'] = len(result['results'])
    return result


def validate_slot_id(slot_id: int):
    if slot_id % 100 > 47:
        raise HTTPException(
            status_code=400, detail="slot id hour part must be between 0-47")
    if slot_id // 100 % 100 < 1 or slot_id // 100 % 100 > 31:
        raise HTTPException(
            status_code=400, detail="slot id day part must be between 1-31")
    if slot_id // 10000 % 100 < 1 or slot_id // 10000 % 100 > 12:
        raise HTTPException(
            status_code=400, detail="slot id month part must be between 1-12")
    if slot_id // 1000000 < 2020 or slot_id // 1000000 > 4000:
        raise HTTPException(
            status_code=400, detail="slot id year part must be resonable")



import os
import jwt
from configparser import ConfigParser


class Auth0Exception(Exception):
    """Exception for auth0"""


class IllegalTokenExcpetion(Auth0Exception):
    """Illegal Auth0 Exception"""


class Auth0TokenVerifier():
    """Does all the token verification using PyJWT"""

    def __init__(self, permissions=None, scopes=None):
        self.permissions = permissions
        self.scopes = scopes
        self.config = {
            "DOMAIN": os.getenv("DOMAIN", "bdating-consumer-dev.au.auth0.com"),
            "API_AUDIENCE": os.getenv("API_AUDIENCE", "oa1z7FUmEYNNv2Bhi8Rk2I6ISO0AdbEN"),
            "ISSUER": os.getenv("ISSUER", "https://bdating-consumer-dev.au.auth0.com/"),
            "ALGORITHMS": os.getenv("ALGORITHMS", "RS256"),
        }

        # This gets the JWKS from a given URL and does processing so you can
        # use any of the keys available
        jwks_url = f'https://{self.config["DOMAIN"]}/.well-known/jwks.json'
        self.jwks_client = jwt.PyJWKClient(jwks_url)

    def verify(self, token):
        # This gets the 'kid' from the passed token
        try:
            signing_key = self.jwks_client.get_signing_key_from_jwt(
                token
            )
            signing_key = signing_key.key
        except jwt.exceptions.PyJWKClientError as error:
            return {"status": "error", "msg": error.__str__()}
        except jwt.exceptions.DecodeError as error:
            return {"status": "error", "msg": error.__str__()}

        try:
            payload = jwt.decode(
                token,
                signing_key,
                algorithms=self.config["ALGORITHMS"],
                audience=self.config["API_AUDIENCE"],
                issuer=self.config["ISSUER"],
            )
        except Exception as e:
            return {"status": "error", "message": str(e)}

        if self.scopes:
            result = self._check_claims(
                payload, 'scope', str, self.scopes.split(' '))
            if result.get("error"):
                return result

        if self.permissions:
            result = self._check_claims(
                payload, 'permissions', list, self.permissions)
            if result.get("error"):
                return result

        return payload

    def _check_claims(self, payload, claim_name, claim_type, expected_value):

        instance_check = isinstance(payload[claim_name], claim_type)
        result = {"status": "success", "status_code": 200}

        payload_claim = payload[claim_name]

        if claim_name not in payload or not instance_check:
            result["status"] = "error"
            result["status_code"] = 400

            result["code"] = f"missing_{claim_name}"
            result["msg"] = f"No claim '{claim_name}' found in token."
            return result

        if claim_name == 'scope':
            payload_claim = payload[claim_name].split(' ')

        for value in expected_value:
            if value not in payload_claim:
                result["status"] = "error"
                result["status_code"] = 403

                result["code"] = f"insufficient_{claim_name}"
                result["msg"] = (f"Insufficient {claim_name} ({value}). You "
                                 "don't have access to this resource")
                return result
        return result



import boto3
from botocore.exceptions import ClientError
import json
import base64
import os

def get_secret(secret_name: str, logger=None) -> dict:
  session = boto3.session.Session()
  client = session.client(
      service_name='secretsmanager',
      region_name=os.environ.get('AWS_REGION', 'ap-southeast-2')
  )
  try:
    get_secret_value_response = client.get_secret_value(
        SecretId=secret_name
    )
  except ClientError as e:
    if logger:
      logger.error(f"Secret: {secret_name} failed, reason: {str(e)}")
    raise e
  else:
    if 'SecretString' in get_secret_value_response:
      secret = get_secret_value_response['SecretString']
    else:
      secret = base64.b64decode(get_secret_value_response['SecretBinary'])
  return json.loads(secret)



#!/usr/bin/env python3
from typing import Dict, Callable
import os
import time
import json
import logging
from bdating_common.collect_default import collect_info_for_apps
#from bdating_common.json_logger import setup_logger

app_info = collect_info_for_apps()
sns_client = app_info['session'].client(
  'sns', region_name=app_info.get('aws_region'))
sqs_client = app_info['session'].client(
  'sqs', region_name=app_info.get('aws_region'),
  endpoint_url=f"https://sqs.{app_info.get('aws_region')}.amazonaws.com")


def report_event(
    _type: str,
    action: str,
    body: Dict[str, object],
    success: bool = True,
    source_id: str = None,
    fail_on_error: bool = False
):
    """Report event
    """
    log = logging.getLogger(__name__)
    log.debug(
        f"Event ({_type}, {action}) to be reported to {app_info.get('event_bus_topic')} with body {body}")
    try:
        sns_client.publish(
            TopicArn=app_info.get('event_bus_topic'),
            Message=json.dumps({
                'time': int(time.time()) * 1000,
                'type': _type,
                'source_id': source_id,
                'action': action,
                'body': body,
                'result': {
                    'success': success,
                },
            }),
            MessageAttributes={
                'type': {
                    'DataType': 'String',
                    'StringValue': _type,
                },
                'action': {
                    'DataType': 'String',
                    'StringValue': action,
                },
            },
        )
    except Exception as e:
        log.warning(e)
        if fail_on_error:
            raise e
    log.info(
        f"Event ({_type}, {action}) reported to {app_info.get('event_bus_topic')} ")


def consume_event_loop(queue: str, process_method: Callable):
    """A skeleton for event consumer.

    Args:
        queue (_type_): the event queue, can be name or simply queue url
        process_method (_type_): _description_
    """
    log = logging.getLogger(__name__)
    if queue.startswith("https"):
        queue_url = queue
    else:
        response = sqs_client.list_queues(
            QueueNamePrefix=queue,
        )
        queue_url = response['QueueUrls'][0]
    error_count = 0
    while True:
        try:
            log.debug("Retrieving event...")
            response = sqs_client.receive_message(
                QueueUrl=queue_url,
                AttributeNames=[
                  'SentTimestamp'
                ],
                MaxNumberOfMessages=1,
                MessageAttributeNames=[
                  'All'
                ],
                WaitTimeSeconds=20
            )
            log.debug(f"Received event count {len(response.get('Messages', []))}")
            if 'Messages' in response:
                for message in response['Messages']:
                    try:
                        process_method(json.loads(message['Body']))
                        response = sqs_client.delete_message(
                            QueueUrl=queue_url,
                            ReceiptHandle=message['ReceiptHandle']
                        )
                        error_count = 0
                    except Exception as e:
                        # pub back.
                        log.warning(f"Failed to process message {message}, reason {e}")
                        sqs_client.change_message_visibility(
                            QueueUrl=queue_url,
                            ReceiptHandle=message['ReceiptHandle'],
                            VisibilityTimeout=2
                        ) 
        except Exception as e:
            log.warning(e)
            error_count += 1
            if error_count > 10:
                time.sleep(4)


if __name__ == '__main__':
    for _ in range(10):
        report_event(
            'test', 'test_action',
            body={"demo": f"this is a demo"},
        )



#!/usr/bin/env python3
import click
import click_log
import logging
from bdating_common import collect_default

log = logging.getLogger(__name__)
click_log.basic_config(log)

# This global info might be used with multiple threading enviroment.
# Think careful of concurrency.
global_info = {
  'session_info':None
} 
def _generate_dict(dimensions):
  result = []
  for kv in dimensions:
    key, value = kv.split('=', 1)
    result.append({
        'Name': key,
        'Value': value
    })

  return result


def _check_int(s:str):
  if s[0] in ('-', '+'):
    return s[1:].isdigit()
  return s.isdigit()

# @multi_threading.threaded
def report_metric(
    namespace: str,
    metric_name: str,
    value, # int or float
    unit: str = 'None',
    aws_profile: str = None,
    aws_region: str = None,
    dimensions: tuple = (),
    resolution: int = 60,
    logger = None
):
  """Report cloudwatch metrics.

  Args:
      namespace ([type], optional): [description].
      metric_name ([type], optional): [description].
      aws_profile ([type], optional): [description]. Defaults to None.
      aws_region ([type], optional): [description]. Defaults to None.
      value ([type], optional): [description].
      dimensions ([type], optional): [description]. Defaults to empty list.
      unit ([type], optional): [description]. Defaults to None. selection: 'Seconds' | 'Microseconds' | 'Milliseconds' | 'Bytes' | 'Kilobytes' | 'Megabytes' | 'Gigabytes' | 'Terabytes' | 'Bits' | 'Kilobits' | 'Megabits' | 'Gigabits' | 'Terabits' | 'Percent' | 'Count' | 'Bytes/Second' | 'Kilobytes/Second' | 'Megabytes/Second' | 'Gigabytes/Second' | 'Terabytes/Second' | 'Bits/Second' | 'Kilobits/Second' | 'Megabits/Second' | 'Gigabits/Second' | 'Terabits/Second' | 'Count/Second' | 'None',
  """
  global global_info
  if not global_info['session_info']:
    global_info['session_info'] = collect_default.collect_info_for_apps(
        aws_profile=aws_profile,
        aws_region=aws_region,
        logger=logger,
    )
  sys_info=global_info['session_info']
  client = sys_info['session'].client('cloudwatch', region_name=sys_info.get('aws_region'))
  if logger:
    logger.debug(f"Metrics to be put to namspace {namespace}, metric name {metric_name}, value {value}")
  client.put_metric_data(
      Namespace=namespace,
      MetricData=[
          {
              'MetricName': metric_name,
              'Dimensions': dimensions,
              # 'Timestamp': datetime.datetime.now(),
              'Value': value,
              'Unit': unit,
              'StorageResolution': resolution
          },
      ]
  )
  if logger:
    logger.info(f"Metrics put for namspace {namespace}, metric name {metric_name}, value {value}")

@click.command()
@click.option('--namespace', '-n', help='The namespace to report', required=True, default='Testing')
@click.option('--profile', '-p', help='The profile', default=None)
@click.option('--region', '-r', help='The AWS region', default='ap-southeast-2')
@click.option('--metric-name', '-m', help='Them metric name', default='Testing')
@click.option('--value', '-V', help='Them metric value', default='1')
@click.option('--unit', '-u', help='Them metric unit', default='None', type=click.Choice([
    'Seconds', 'Microseconds', 'Milliseconds', 'Bytes', 'Kilobytes', 'Megabytes', 'Gigabytes',
    'Terabytes', 'Bits', 'Kilobits', 'Megabits', 'Gigabits', 'Terabits', 'Percent', 'Count',
    'Bytes/Second', 'Kilobytes/Second', 'Megabytes/Second', 'Gigabytes/Second', 'Terabytes/Second',
    'Bits/Second', 'Kilobits/Second', 'Megabits/Second', 'Gigabits/Second', 'Terabits/Second',
    'Count/Second' , 'None',
],))
@click.option('--dimension', '-d', help='The Dimention to add. Format KEY=value.',
              default=[], multiple=True)
@click_log.simple_verbosity_option(log)
def report(namespace, profile, region, metric_name, value, unit, dimension):
  """Report a metric"""
  report_metric(
      namespace=namespace,
      aws_profile=profile,
      aws_region=region,
      metric_name=metric_name,
      value=int(value) if _check_int(value) else float(value),
      dimensions=_generate_dict(dimension),
      unit=unit,
      logger=log,
  )
  if not namespace == 'Testing' and not namespace.startswith('Sonder/'):
    namespace = f"Sonder/{namespace}"
  log.info(f"https://{region}.console.aws.amazon.com/cloudwatch/home?region={region}#metricsV2:graph=~(view~'timeSeries~stacked~false~region~'ap-southeast-2~stat~'Minimum~period~60);query=~'{namespace}")

if __name__ == '__main__':
  report()



"""
Notification Helper

For each user (id is wallet) and _type, system push notifications to the user.
Once this user reads his notification, the notification will be gone from system. The notification can be read only once.
The notificaiton has a default TTL to 1 hour.

"""

from typing import Dict, Any, List
import redis.asyncio as redis
from time import time, sleep

from bdating_common.api_common import settings
redis_client = redis.Redis(
    host=settings.redis_host,
    port=settings.redis_port,
    db=settings.in_app_notif_cache_id,
    password=settings.redis_password
)


async def push(wallet: str, _type: str, message: str, timestamp: Any = None, ttl: int = 3600):
    """TTL in milli-seconds
    """
    if timestamp is None:
        timestamp = int(time())
    await redis_client.setex(
        f"{wallet}:{_type}:{timestamp}", ttl, message
    )


async def pop(wallet: str, _type: str, delete: bool = True) -> List[Dict[str, str]]:
    """Return notification for this user, ordered by timestamp
    If not exist, return empty list.

    when _type is set to None, it will get all notifications for a wallet.
    """
    search_key = f"{wallet}:"
    if _type is not None:
        search_key = f"{search_key}{_type}:"

    keys = (await redis_client.scan(
        cursor=0, match=f"{search_key}*"
    ))[-1]
    if not keys:
        return []
    else:
        result = []
        keys.sort()

        for obj_key in keys:

            if delete:
                r = await redis_client.getdel(obj_key)
            else:
                r = await redis_client.get(obj_key)
            if r is not None:  # concurreny protection
                result.append(r.decode('utf8'))

        return [r for r in result if r is not None]


async def _local_test():
    await push('3', 'consumer', "hello222", 222, 12)
    await push('3', 'provider', "hello221", 221, 12)
    await push('3', 'consumer', "hello231", 231, 12)
    await push('3', 'provider', "hello232", 232, 12)
    sleep(2)
    assert ['hello222', 'hello231'] == (await pop('3', 'consumer', delete=False))
    assert ['hello221', 'hello232'] == (await pop('3', 'provider', delete=False))
    # order by type first.
    assert ['hello222', 'hello231', 'hello221', 'hello232'] == (await pop('3', None, delete=False))
    assert ['hello222', 'hello231'] == (await pop('3', 'consumer', delete=True))
    assert ['hello221', 'hello232'] == (await pop('3', None, delete=False))

    await push('3', 'consumer', "hello222", 222, 12)
    await push('3', 'provider', "hello221", 221, 12)
    await push('3', 'consumer', "hello231", 231, 12)
    await push('3', 'provider', "hello232", 232, 12)
    assert ['hello222', 'hello231'] == (await pop('3', 'consumer'))
    assert ['hello221', 'hello232'] == (await pop('3', 'provider'))

    assert [] == (await pop('3', 'consumer'))
    assert [] == (await pop('3', 'provider'))
    assert [] == (await pop('3', None))

    await push('3', 'consumer', "hello222", 222, 12)
    sleep(20)
    assert [] == (await pop('3', None))

if __name__ == "__main__":
    import asyncio
    asyncio.run(_local_test())



import logging
from typing import Tuple
from datetime import datetime
log = logging.getLogger(__name__)
MIN_SLOT_IN_DAY = 0
MAX_SLOT_IN_DAY = 47



def parse_slot(slot_id: int) -> Tuple[datetime, int]:

    try:
        slot = slot_id % 100
        if slot < MIN_SLOT_IN_DAY or slot > MAX_SLOT_IN_DAY:
            return None, None
        return datetime(slot_id // 1000000, slot_id // 10000 % 100, slot_id // 100 % 100), slot
    except ValueError:
        return None, None


def convert_slot_to_epoch(slot_id: int) -> int:
    date_time, _ = parse_slot(slot_id=slot_id)
    return int(date_time.timestamp()) + slot_id % 100 * 30 * 60



if __name__ == "__main__":
    print(convert_slot_to_epoch(2022010123))


import os
def is_env_true(env_key:str):
    env_value=os.environ.get(env_key)
    if env_value is None:
        return False
    env_value=env_value.strip().lower()
    if env_value.startswith('t') or env_value.startswith('y'):
        return True
    if env_value.isdigit():
        return False if env_value.replace('0', '')=='' else True # all constructed with 0
    return False


"""
Crypto helper.

"""
import logging
from tokenize import Number
import web3
from web3 import Web3, HTTPProvider
from web3.middleware import geth_poa_middleware
# from web3.auto.infura.mainnet import w3
from web3.exceptions import TransactionNotFound
import pickle
import sys
from enum import Enum
import json
import requests

logger = logging.getLogger(__name__)

class TxnStatusEnum(Enum):
  TXN_MISMATCHED = -2
  TXN_REVERTED = -1
  TXN_NOT_READY = 0
  TXN_CONFIRMED = 1


networks = {
  'bnb_test': {
    # 'rpc_url': 'https://data-seed-prebsc-1-s1.binance.org:8545/',
    # 'chain_id': '97',
    # 'currency': 'BNB',
    # 'explorer_url': 'https://testnet.bscscan.com',
    'w3_client': Web3(HTTPProvider('https://data-seed-prebsc-1-s1.binance.org:8545/')),
    'minimal_confirm': 2
  },
  'bnb_main': {
    # 'rpc_url': 'https://bsc-dataseed1.ninicoin.io',
    # 'chain_id': '56',
    # 'currency': 'BNB',
    # 'explorer_url': 'https://bscscan.com',
    'w3_client': Web3(HTTPProvider('https://bsc-dataseed1.ninicoin.io')),
    'minimal_confirm': 3
  },
  'eth_main': {
    # 'rpc_url': 'https://rpc.ankr.com/eth',  # https://rpc.info/
    # 'chain_id': '1',
    # 'currency': 'ETH',
    'explorer_url': 'https://rpc.ankr.com/eth',
    'w3_client': Web3(HTTPProvider('https://rpc.ankr.com/eth')),
    'minimal_confirm': 3
  },
  'eth_goerli': {
    'rpc_url': 'https://rpc.ankr.com/eth_goerli',
    'chain_id': '5',
    'currency': 'ETH',
    'explorer_url': 'https://goerli.etherscan.io',
    'w3_client': Web3(HTTPProvider('https://rpc.ankr.com/eth_goerli')),
    'minimal_confirm': 3,
    'contract_address': '0xB99d86242bb33F874F1Eb59dd0C403369a7D1D7c'
  }
}
for k, v in networks.items():
    if k.endswith(''):
        v['w3_client'].middleware_onion.inject(geth_poa_middleware, layer=0)

""" Depricated
def check_transactions_by_wallet(blockchain_address: str, network: str = 'bnb_main', Logger: object=None):
    # NOTE: WIP code. We are not sure when and how to use this method.
    w3 = networks[network]['w3_client']

    # request the latest block number
    ending_blocknumber = w3.eth.blockNumber

    # latest block number minus 100 blocks
    starting_blocknumber = ending_blocknumber - 1

    # create an empty dictionary we will add transaction data to
    tx_dictionary = {}

    print(
        f"Started filtering through block number {starting_blocknumber} to {ending_blocknumber} for transactions involving the address - {blockchain_address}...")
    for x in range(starting_blocknumber, ending_blocknumber):
        block = w3.eth.getBlock(x, True)
        for transaction in block.transactions:
            # print(transaction)
            if str(transaction['to'].lower()) == blockchain_address or str(transaction['from'].lower()) == blockchain_address:
                print(transaction)
                with open("transactions.pkl", "wb") as f:
                    hashStr = transaction['hash'].hex()
                    tx_dictionary[hashStr] = transaction
                    pickle.dump(tx_dictionary, f)
                f.close()
    print(f"Finished searching blocks {starting_blocknumber} through {ending_blocknumber} and found {len(tx_dictionary)} transactions")
"""

def check_transaction_status(transaction_hash: str, to_address: str, value: Number, chain_id: str = None, network: str = 'bnb_main') -> TxnStatusEnum:
    if chain_id is not None:
      for name, network_config in networks.items():
        if network_config.get('chain_id') == chain_id:
          network = name
    w3 = networks[network]['w3_client']
    print(f"check_transaction_status(), trying to check the following transaction on network {network}")
    print(f"check_transaction_status(), transaction hash: {transaction_hash}")
    print(f"check_transaction_status(), expected: token_contract={networks[network]['contract_address']}")
    print(f"check_transaction_status(), expected: token_to={to_address}")
    print(f"check_transaction_status(), expected: tokens_value={value}")

    minimal_confirm = networks[network]['minimal_confirm']
    try:
        txn_status = w3.eth.get_transaction_receipt(transaction_hash)['status']
        if txn_status == 0:
          logger.warning(f"Transaction {transaction_hash} was reverted by EVM!")
          return TxnStatusEnum.TXN_REVERTED

        transaction = w3.eth.get_transaction(transaction_hash=transaction_hash)

        last_block = w3.eth.get_block_number()
        print(f"Received Tranasction, {transaction}, last_block: {last_block}")
        # print(f"transaction.input= {w3.toBytes(transaction.input)}")
        # abi_endpoint = f'https://api-goerli.etherscan.io/api?module=contract&action=getabi&address=0xB99d86242bb33F874F1Eb59dd0C403369a7D1D7c'
        # abi = json.loads(requests.get(abi_endpoint).text)
        abi = "[{\"constant\":true,\"inputs\":[],\"name\":\"name\",\"outputs\":[{\"name\":\"\",\"type\":\"string\"}],\"payable\":false,\"stateMutability\":\"view\",\"type\":\"function\"},{\"constant\":false,\"inputs\":[{\"name\":\"spender\",\"type\":\"address\"},{\"name\":\"tokens\",\"type\":\"uint256\"}],\"name\":\"approve\",\"outputs\":[{\"name\":\"success\",\"type\":\"bool\"}],\"payable\":false,\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"constant\":true,\"inputs\":[],\"name\":\"totalSupply\",\"outputs\":[{\"name\":\"\",\"type\":\"uint256\"}],\"payable\":false,\"stateMutability\":\"view\",\"type\":\"function\"},{\"constant\":false,\"inputs\":[{\"name\":\"from\",\"type\":\"address\"},{\"name\":\"to\",\"type\":\"address\"},{\"name\":\"tokens\",\"type\":\"uint256\"}],\"name\":\"transferFrom\",\"outputs\":[{\"name\":\"success\",\"type\":\"bool\"}],\"payable\":false,\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"constant\":true,\"inputs\":[],\"name\":\"decimals\",\"outputs\":[{\"name\":\"\",\"type\":\"uint8\"}],\"payable\":false,\"stateMutability\":\"view\",\"type\":\"function\"},{\"constant\":true,\"inputs\":[],\"name\":\"_totalSupply\",\"outputs\":[{\"name\":\"\",\"type\":\"uint256\"}],\"payable\":false,\"stateMutability\":\"view\",\"type\":\"function\"},{\"constant\":true,\"inputs\":[{\"name\":\"tokenOwner\",\"type\":\"address\"}],\"name\":\"balanceOf\",\"outputs\":[{\"name\":\"balance\",\"type\":\"uint256\"}],\"payable\":false,\"stateMutability\":\"view\",\"type\":\"function\"},{\"constant\":true,\"inputs\":[],\"name\":\"symbol\",\"outputs\":[{\"name\":\"\",\"type\":\"string\"}],\"payable\":false,\"stateMutability\":\"view\",\"type\":\"function\"},{\"constant\":true,\"inputs\":[{\"name\":\"a\",\"type\":\"uint256\"},{\"name\":\"b\",\"type\":\"uint256\"}],\"name\":\"safeSub\",\"outputs\":[{\"name\":\"c\",\"type\":\"uint256\"}],\"payable\":false,\"stateMutability\":\"pure\",\"type\":\"function\"},{\"constant\":false,\"inputs\":[{\"name\":\"to\",\"type\":\"address\"},{\"name\":\"tokens\",\"type\":\"uint256\"}],\"name\":\"transfer\",\"outputs\":[{\"name\":\"success\",\"type\":\"bool\"}],\"payable\":false,\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"constant\":true,\"inputs\":[{\"name\":\"a\",\"type\":\"uint256\"},{\"name\":\"b\",\"type\":\"uint256\"}],\"name\":\"safeDiv\",\"outputs\":[{\"name\":\"c\",\"type\":\"uint256\"}],\"payable\":false,\"stateMutability\":\"pure\",\"type\":\"function\"},{\"constant\":false,\"inputs\":[{\"name\":\"spender\",\"type\":\"address\"},{\"name\":\"tokens\",\"type\":\"uint256\"},{\"name\":\"data\",\"type\":\"bytes\"}],\"name\":\"approveAndCall\",\"outputs\":[{\"name\":\"success\",\"type\":\"bool\"}],\"payable\":false,\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"constant\":true,\"inputs\":[{\"name\":\"a\",\"type\":\"uint256\"},{\"name\":\"b\",\"type\":\"uint256\"}],\"name\":\"safeMul\",\"outputs\":[{\"name\":\"c\",\"type\":\"uint256\"}],\"payable\":false,\"stateMutability\":\"pure\",\"type\":\"function\"},{\"constant\":true,\"inputs\":[{\"name\":\"tokenOwner\",\"type\":\"address\"},{\"name\":\"spender\",\"type\":\"address\"}],\"name\":\"allowance\",\"outputs\":[{\"name\":\"remaining\",\"type\":\"uint256\"}],\"payable\":false,\"stateMutability\":\"view\",\"type\":\"function\"},{\"constant\":true,\"inputs\":[{\"name\":\"a\",\"type\":\"uint256\"},{\"name\":\"b\",\"type\":\"uint256\"}],\"name\":\"safeAdd\",\"outputs\":[{\"name\":\"c\",\"type\":\"uint256\"}],\"payable\":false,\"stateMutability\":\"pure\",\"type\":\"function\"},{\"inputs\":[],\"payable\":false,\"stateMutability\":\"nonpayable\",\"type\":\"constructor\"},{\"payable\":true,\"stateMutability\":\"payable\",\"type\":\"fallback\"},{\"anonymous\":false,\"inputs\":[{\"indexed\":true,\"name\":\"from\",\"type\":\"address\"},{\"indexed\":true,\"name\":\"to\",\"type\":\"address\"},{\"indexed\":false,\"name\":\"tokens\",\"type\":\"uint256\"}],\"name\":\"Transfer\",\"type\":\"event\"},{\"anonymous\":false,\"inputs\":[{\"indexed\":true,\"name\":\"tokenOwner\",\"type\":\"address\"},{\"indexed\":true,\"name\":\"spender\",\"type\":\"address\"},{\"indexed\":false,\"name\":\"tokens\",\"type\":\"uint256\"}],\"name\":\"Approval\",\"type\":\"event\"}]" 
        if transaction['to'] != networks[network]['contract_address']:
          print(f"Transaction {transaction_hash} is not using correct contract {networks[network]['contract_address']}.")
          return TxnStatusEnum.TXN_MISMATCHED
        
        contract = w3.eth.contract(address=networks[network]['contract_address'], abi=abi)
        func_obj, func_params = contract.decode_function_input(transaction["input"])
        print(f"transaction.input.func_obj= {func_obj}")
        print(f"transaction.input.func_params= {func_params}")
        
        if last_block - transaction.get('blockNumber', sys.maxsize) >= minimal_confirm:
          if func_params.get('to').lower() == to_address.lower() and func_params.get('tokens') >= value:
            logger.info(f"Transaction {transaction_hash} is available.")
            print(f"Transaction {transaction_hash} is available.")
            return TxnStatusEnum.TXN_CONFIRMED
          else:
            logger.warning(f"Transaction {transaction_hash} has been confirmed but details mismatched!")
            logger.warning(f"Found Transaction: token_to={ func_params.get('to')} , expected: token_to={to_address} ")
            logger.warning(f"Found Transaction: tokens_value={ func_params.get('tokens')} , expected: tokens_value={value} ")
            return TxnStatusEnum.TXN_MISMATCHED
        else:
          logger.info(f"Transaction {transaction_hash} is available now but still needs more confirmation.")
          return TxnStatusEnum.TXN_NOT_READY
    except TransactionNotFound:
        logger.info(f"Transaction {transaction_hash} not found yet.")
        return TxnStatusEnum.TXN_NOT_READY

"""
did not check confidence block gap
"""
def check_transaction_status_v2(transaction_hash: str, to_address: str, value: Number, network: str = 'bnb_main') -> bool:
    w3 = networks[network]['w3_client']
    try:
      transaction = w3.eth.get_transaction(transaction_hash=transaction_hash)
      #TODO: convert bnb value from its unit
      txn_value = Web3.fromWei(transaction['value'], 'ether')
      txn_status = w3.eth.get_transaction_receipt(transaction_hash)['status']
      # status ==1 should be able to present the succeed of txn
      if txn_status == 1:
        if str(transaction['to']).lower() == to_address.lower() and txn_value >= value:
          logger.info(f"Transaction {transaction_hash} is available.")
          print(f"Transaction {transaction_hash} is available.")
          return True
        else:
          logger.warning(f"Transaction {transaction_hash} received but details mismatched. value={txn_value}, to_addr={transaction['to']}")
          return False
      else:
        logger.warning(f"Transaction {transaction_hash} not confidence yet or failed")
        return False


    except TransactionNotFound:
      logger.info(f"Transaction {transaction_hash} not found yet.")
      return False

if __name__ == '__main__':
    # print(check_transaction_status(transaction_hash='0xc878288e8222cc472a450a44ac0674498290757ac101f54b79be05c03d023d51', 
    #   to_address='0x840851d656e2575a3d524af2be7249dcdbaa718c', 
    #   value=0.000001,
    #   network='eth_goerli'))


    # w3 = Web3(HTTPProvider('https://rpc.ankr.com/eth_goerli'));
    # print(w3.eth.filter('pending'))
    # print(Web3.fromWei(677262836600000, 'ether'))

    print(check_transaction_status(transaction_hash='0x11207b48fb51c065adfdab9064c732067cd9a5514b6d33a99c2bc5dbc6163b38', 
      to_address='0x978258905A52De6afFd8F04929204215c1B59A87', 
      value=15000,
      network='eth_goerli')) 




import os
import logging
import json
from inspect import getmembers, isclass
import asyncio

from pydantic import BaseSettings

from elasticsearch import Elasticsearch
from elasticsearch.exceptions import NotFoundError

from fastapi import FastAPI, Depends, status
from fastapi.security import HTTPBearer, OAuth2PasswordBearer
from fastapi.encoders import jsonable_encoder
from fastapi.middleware.cors import CORSMiddleware

from bdating_common.model import HealthResponse, ConsumerProfile, MultiLangEnum
from bdating_common import model as model_module
# from bdating_common import model_zh as model_zh_module
# from bdating_common import model_en as model_en_module

from bdating_common.auth0_token_helper import IllegalTokenExcpetion, Auth0TokenVerifier
from fastapi.responses import JSONResponse
from bdating_common.es_helper import es_get_result_to_dict
from fastapi import WebSocket
from fastapi.websockets import WebSocket, WebSocketDisconnect
import redis
import aioredis
from aioredis.pubsub import Channel
from auth0.v3.authentication import Users as Auth0Users
from typing import Union, Any, Dict

oauth2_scheme = OAuth2PasswordBearer(tokenUrl='token')

log = logging.getLogger(__name__)


def get_model_enum_values():
    model_enums = {}
    for name, member in getmembers(model_module, isclass):
        if issubclass(member, MultiLangEnum) and member is not MultiLangEnum:
            try:
                trans = member.translates()
                res = []
                for key in member:
                    _d = {'_key': key}
                    for lang in ('en', 'zh'):
                        # if the given key is not need to be translated
                        if key not in trans:
                            _d[lang] = key.replace('_', ' ').capitalize()
                        else:
                            _d[lang] = trans[key].get(lang)
                            if _d[lang] is None:
                                _d[lang] = key.replace('_', ' ').capitalize()
                    res.append(_d)
                model_enums[name] = res
            except (AttributeError, KeyError) as ex:
                log.warning('Enum %s has no proper translation defined', name)
                res = []
                for m in member.__members__:
                    res.append({'_key': m, 'en': m.replace(
                        '_', ' ').capitalize(), 'zh': m.replace('_', ' ').capitalize()})
                model_enums[name] = res
    return model_enums


"""
def get_model_enum_values2():
    model_enums = {}
    for data_type_name in dir(model_module):
        if not data_type_name.startswith('__') and data_type_name != 'Enum' and type(model_module.__dict__[data_type_name]) == enum.EnumMeta:
            model_enums[data_type_name] = [
                {
                    '_key': variable_name,
                    'en': model_en_module.__dict__[data_type_name].__dict__[variable_name].value,
                    'zh': model_zh_module.__dict__[data_type_name].__dict__[variable_name].value,
                }
                for variable_name in dir(model_module.__dict__[data_type_name]) if not variable_name.startswith('__')
            ]
    return model_enums
"""


class Settings(BaseSettings):
    app_name: str = 'StandardBdatingAPI'
    app_namespace: str = os.getenv('NAMESPACE', "")
    admin_email: str = "admin@bdating.io"
    app_type: str = 'consumer'
    es_endpoint: str = os.getenv('ELASTICSEARCH_HOSTS')
    es_index: str = 'bdating'
    auth0_namespace: str = 'https://app.bdating.io/'
    redis_host: str = os.getenv('REDIS_HOST')
    redis_port: int = 6379
    redis_password: str = os.getenv('REDIS_PASSWORD')
    cors_origins: str = os.getenv('CORS_ORIGINS', 'http://localhost:3000')
    token_cache_db_id: int = 0
    lock_db_id: int = 1
    in_app_notif_cache_id: int = 2


settings = Settings()
token_auth_scheme = HTTPBearer()
token_verifier = Auth0TokenVerifier()

token_cache_redis = None  # TODO rework the global clients.


def _get_uid(token: str):
    global token_cache_redis
    if settings.redis_host:
        if token_cache_redis is None:
            token_cache_redis = redis.Redis(
                settings.redis_host, settings.redis_port,
                db=settings.token_cache_db_id,
                password=settings.redis_password
                )
        cached_value = token_cache_redis.get(token)
        if cached_value is not None:
            return json.loads(cached_value.decode('utf-8')).get(f"{settings.auth0_namespace}uid")
    try:
        payload = token_verifier.verify(token)
        if payload.get("status") == 'error':
            raise IllegalTokenExcpetion()
        uid = payload.get(f"{settings.auth0_namespace}uid")
        log.warn('uid retrieved is %s', uid)
        if token_cache_redis is not None and uid:
            ttl = 86400 * 1000  # hard code 1 day, ttl value is in milliseconds
            token_cache_redis.setex(
                token, ttl, json.dumps(payload, default=str))
        return uid
    except Exception as e:
        log.warn('Token verification failed', exc_info=True)
        raise IllegalTokenExcpetion(e)


async def validate_socket_token(token: str, websocket: WebSocket):
    global token_cache_redis
    if settings.redis_host:
        if token_cache_redis is None:
            token_cache_redis = redis.Redis(
                settings.redis_host, settings.redis_port,
                db=settings.token_cache_db_id,
                password=settings.redis_password
                )
        cached_value = token_cache_redis.get(token)
        if cached_value is not None:
            payload = json.loads(cached_value.decode('utf-8'))
            return payload.get(f"{settings.auth0_namespace}wallet", payload.get('wallet'))
    try:
        payload = token_verifier.verify(token)
        if payload.get("status") == 'error':
            raise IllegalTokenExcpetion(payload)
        wallet = payload.get(
            f"{settings.auth0_namespace}wallet", payload.get('wallet'))
        log.warn('wallet retrieved is %s', wallet)
        if token_cache_redis is not None and wallet:
            ttl = 86400 * 1000  # hard code 1 day, ttl value is in milliseconds
            token_cache_redis.setex(
                token, ttl,  json.dumps(payload, default=str))
        return wallet
    except Exception as ex:
        log.exception(ex)
        await websocket.send_text('{"status":"403"}')
        await websocket.close()
        raise IllegalTokenExcpetion(ex)


def get_uid(token: str, token_verifier: object):
    global token_cache_redis
    if settings.redis_host:
        if token_cache_redis is None:
            token_cache_redis = redis.Redis(
                settings.redis_host, settings.redis_port,
                db=settings.token_cache_db_id,
                password=settings.redis_password
                )
        cached_value = token_cache_redis.get(token.credentials)
        if cached_value is not None:
            return cached_value.decode('utf-8')
    try:
        result = token_verifier.verify(token.credentials)
        log.warn('result %s', result)
        if result.get("status") == 'error':
            raise IllegalTokenExcpetion()
        domain = os.environ.get('DOMAIN')
        log.warn('domain %s', domain)
        users = Auth0Users(domain)
        userinfo = users.userinfo(token.credentials)
        log.warn('userinfo %s', userinfo);
        uid = userinfo.get(f"{settings.auth0_namespace}uid")
        # value = result.get(f"{settings.auth0_namespace}uid")
        # log.warn('uid retrieved is %s', value)
        if token_cache_redis is not None and uid:
            ttl = 86400 * 1000  # hard code 1 day, ttl value is in milliseconds
            token_cache_redis.setex(token.credentials, ttl, uid)
        return uid
    except Exception as e:
        log.warn('Token verification failed', exc_info=True)
        raise IllegalTokenExcpetion(e)


def get_contact(token: str):
    try:
        payload = token_verifier.verify(token)
        if payload.get('status') == 'error':
            raise IllegalTokenExcpetion()
        contact = payload.get(f"{settings.auth0_namespace}name")
        return contact
    except Exception as e:
        log.warn('Token verification failed', exc_info=True)
        raise IllegalTokenExcpetion(e)


async def ws_reader(channel: Channel, ws: WebSocket):
    while await channel.wait_message():
        text = None
        try:
            text = (await channel.get()).decode('utf-8')
            msg = json.loads(text)
            await ws.send_text(json.dumps({**msg, "channel": channel.name.decode('utf-8')}, default=str))
        except Exception as e:
            log.error(
                'Could not process channel mesage from %s. Message: %s. Error: e', channel.name, text, e)


async def redis_connector(
    websocket: WebSocket, wallet: str
):
    async def consumer_handler(ws: WebSocket, redis_client: object):
        try:
            own_channel = (await redis_client.subscribe(wallet))[0]
            asyncio.create_task(ws_reader(own_channel, ws))
            additional_channel = None
            while True:
                message = await ws.receive_text()
                message = message.strip()
                if message.lower() == "close":
                    await websocket.close()
                    break
                else:
                    if additional_channel:
                        redis_client.unsubscribe(additional_channel)
                    additional_channel = (await redis_client.subscribe(message))[0]
                    asyncio.create_task(ws_reader(additional_channel, ws))
        except WebSocketDisconnect:
            log.debug("Socket disconnected, no big deal")

    redis = await aioredis.create_redis_pool(
        address=f"redis://{settings.redis_host}:{settings.redis_port}",
        db=settings.in_app_notif_cache_id,
        password=settings.redis_password
    )
    consumer_task = consumer_handler(websocket, redis)
    done, pending = await asyncio.wait(
        [consumer_task, ], return_when=asyncio.FIRST_COMPLETED,
    )
    log.debug(f"Done task: {done}")
    for task in pending:
        log.warning(f"Canceling task: {task}")
        task.cancel()
    redis.close()
    await redis.wait_closed()


def create_app():
    app = FastAPI(title=settings.app_name)
    es = Elasticsearch(settings.es_endpoint)

    @ app.get("/health", response_model=HealthResponse)
    def get_health():
        return {"status": "OK"}

    @app.exception_handler(NotFoundError)
    async def es_not_found_handler(request, exc):
        return JSONResponse(
            status_code=status.HTTP_404_NOT_FOUND,
            content=jsonable_encoder({"detail": "Not found"}),)

    @app.exception_handler(IllegalTokenExcpetion)
    async def illegal_Auth0_token_exception_handler(request, exc):
        return JSONResponse(
            status_code=status.HTTP_403_FORBIDDEN,
            content=jsonable_encoder({"detail": "Illegal token"}),)

    @app.get("/me")
    def get_own_profile_info(token: str = Depends(token_auth_scheme)):
        """
        Find this user's own profile info.
        """
        uid = get_uid(token, token_verifier)
        return es_get_result_to_dict(es.get(index=settings.es_index, id=f"{uid}:{settings.app_type}"))

    @app.patch("/me")
    def update_own_profile_info(profile: ConsumerProfile, token: str = Depends(token_auth_scheme)):
        """
        Update the profile for the user.
        If the profile does not exist, create it in the first place.
        """

        profile = jsonable_encoder(profile)
        log.debug('profile %s', profile)
        log.debug('token %s', token)
        uid = get_uid(token, token_verifier)
        return {}
        return es.update(index=settings.es_index, id=f"{uid}:{settings.app_type}", doc_as_upsert=True, doc=profile)

    @app.get('/config/app')
    def get_app_config():
        """get application default configuration"""
        result = {
            'enum_values': get_model_enum_values()
        }
        return result

    if settings.cors_origins:
        origins = settings.cors_origins.split(',')
        app.add_middleware(
            CORSMiddleware,
            allow_origins=origins,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
    return app


def publish_message(message: Union[Dict[Any, Any], str], wallet: str, sync_redis: object = None):
    """Publish a message to a user identified by wallet"""
    if not sync_redis:
        sync_redis = redis.Redis(
            host=settings.redis_host,
            port=settings.redis_port,
            db=settings.in_app_notif_cache_id,
            password=settings.redis_password,
        )
    if isinstance(message, Dict):
        sync_redis.publish(wallet, json.dumps(message, default=str))
    else:
        sync_redis.publish(wallet, json.dumps({
            "message": message
            }))



from fastapi.responses import JSONResponse

GENERAL_ERROR = 'GENERAL_ERROR'
GENERAL_ERROR_MESSAGE = 'Something is wrong.'

def create_error(code: str=GENERAL_ERROR, message: str=GENERAL_ERROR_MESSAGE, exception: Exception = None):
    if exception is not None:
        message = message + ' (Exception: ' + str(exception) + ')'
    return {
        'error' : {
            'code': code,
            'message': message
        }
    }

class BaseError(JSONResponse):
    pass

class BookingNotFound(BaseError):
    code = 'BOOKING_NOT_FOUND'
    message = 'Booking id {BOOKING_ID} does not exists.'
    def __init__(self, booking_id=None):
        super().__init__(content={
            'error': {
                'code': self.code,
                'message': self.message,
                'BOOKING_ID': booking_id
            }
        })

class BookingStatusError(BaseError):
    code = 'BOOKING_STATUS_ERROR'
    message = 'Action cannot be performed on booking {BOOKING_ID} when its status is {BOOKING_STATUS}.'
    def __init__(self, booking_id=None, booking_status=''):
        super().__init__(content={
            'error': {
                'code': self.code,
                'message': self.message,
                'BOOKING_ID': booking_id,
                'BOOKING_STATUS': booking_status,
            }
        })

class BookingTimeout(BaseError):
    code = 'BOOKING_TIMEOUT'
    message = 'Request timed out, please try again later.'
    def __init__(self):
        super().__init__(content={
            'error': {
                'code': self.code,
                'message': self.message,
            }
        })

class BookingRequestError(BaseError):
    code = 'BOOKING_REQUEST_ERROR'
    message = 'Booking request error.'
    def __init__(self, ex=None):
        super().__init__(content={
            'error': {
                'code': self.code,
                'message': self.message,
                'exception': str(ex)
            }
        })



"""
Scheduler DB entry Helper for Bdating.
"""
from elasticsearch import Elasticsearch, NotFoundError
import os
import logging
from bdating_common.es_helper import es_search_result_to_dict
from bdating_common.api_common import settings
from bdating_common.event_helper import report_event
settings.es_index = f"{settings.app_namespace}-bdating"
es = Elasticsearch(settings.es_endpoint)

whoami = os.getenv('WHO_AM_I', 'bdating-common-library')
logger = logging.getLogger(whoami)
logger.setLevel(os.getenv("LOG_LEVEL", 'INFO'))


def find_scheduled_items(now: int, max_size: int = 1000):
    """
    Find all qualified alarms to send, and then process them
    1. Find all records start time < now, then process them and generate events.
    1. Find all records no_later_than < now then delete them.

    """
    query_condition = {
        "query": {
            "bool": {
                "must": [
                    {
                        "term": {
                            "type": "schedule"
                        }
                    },
                    {
                        "term": {
                            "status": "active"
                        }
                    },
                    {
                        "range": {
                            "start_time": {
                                "lte": now
                            }
                        }
                    },
                    # {
                    #     "range": {
                    #         "no_later_than": {
                    #             "gte": now
                    #         }
                    #     }
                    # }
                ]
            }
        },
        "sort": [
            {"start_time": "asc"}
        ],
        "size": max_size
    }

    return es_search_result_to_dict(es.search(index=settings.es_index, body=query_condition), with_id=True)


def create_schedule(schedule_type: str, matching_id: str, start_time: int, no_later_than: int = None, **args):
    """
    Args: the definition of args are following the schedule_type
    """
    if not no_later_than:
        no_later_than = start_time+300  # must finish within 5 mins as default setting
    record = {
        "type": "schedule",
                "schedule_type": schedule_type,  # notify, verify_payment, etc.
                # booking id/transaction id, or any other uniq id that can decide the related schedules
                "matching_id": matching_id,
                "status": "active",  # place holder. Always active.
                "start_time": start_time,  # started to execute after this time
                # if now is already after this time, then cancel this.
                "no_later_than": no_later_than,
                "body": args
    }
    es.index(index=settings.es_index, document=record)
    logger.debug(f"Report event: schedules for id {matching_id} are created.")
    report_event(
        _type='schedule',
        action='created',
        body={
            'document': record
        },
        success=True,
        source_id=whoami,
    )


def cancel_schedule(schedule_or_matching_id: object):
    try:
        if isinstance(schedule_or_matching_id, dict):  # expect to be scheudle
            record_id = schedule_or_matching_id.get('_id')
            logger.debug(f"Cancel schedule by its record_id {record_id}.")
            es.delete(
                index=settings.es_index,
                id=record_id
            )
        else:
            matching_id = schedule_or_matching_id
            logger.debug(f"Cancel schedule by its original id {matching_id}.")
            es.delete_by_query(
                index=settings.es_index,
                query={
                    "match": {
                        "matching_id": matching_id
                    }
                })
    except Exception as e:
        if not isinstance(e, NotFoundError):
            logger.warning(
                f"Exception when deleting related records for {schedule_or_matching_id}. Excepti")
    #logger.info(f"Schedules mathcing {schedule_or_matching_id} are deleted.")

    #logger.debug(
    #    f"Report event: schedules related to {schedule_or_matching_id} are canceled.")
    report_event(
        _type='schedule',
        action='canceled',
        body={
            'matching_condition': schedule_or_matching_id
        },
        success=True,
        source_id=whoami,
    )



import logging
import os
import requests
import json

log = logging.getLogger(__name__)


def notify(recipient: str, message: str, wait: bool = False, retry_times: int = 3) -> bool:
    """Send Message to recipient. If recipient looks like a supported mobile number, then send SMS, otherwise if looks like a email, then send email. If wait, then wait until send is successful or fail after retry.

    Return true if sent successfully. Otherwise return false.
    """
    # TODO: add async and retry on failure.
    try:
        if '@' not in recipient and not recipient.startswith("+"):
            log.error(
                f"The recipient {recipient} does not look like a email or sms. Giving up.")
            return False

        resp = requests.post(
            url=os.environ['NOTIFY_URL'],
            data=json.dumps({
                "recipient": recipient,
                "body": message,
                "sender": os.environ['NOTIFY_TOKEN']
            })
        )
        if resp.status_code//100 == 2:
            return True
        else:
            log.error(
                f"Failed to send notification. {resp.status_code}, {resp.text}")
            return False
    except Exception as e:
        log.error(
            f"Failed to send {message[0:20]} to {recipient}. Exception {e}. Giving up.")
        return False


if __name__ == "__main__":
    assert True == notify(
        recipient='help@bdating.io',
        message='Please visit https://www.bdating.io')
    assert True == notify(
        recipient='+61431009880',
        message='Please visit https://www.bdating.io')
    assert True == notify(
        recipient='+6143', # this library does not care
        message='Please visit https://www.bdating.io')
    assert False == notify(
        recipient='abcde',
        message='Please visit https://www.bdating.io')



import threading
from functools import wraps
def threaded(f):
    ''' This decorator executes a function in a Thread'''
    @wraps(f)
    def wrapper(*args, **kwargs):
        thr = threading.Thread(target=f, args=args, kwargs=kwargs)
        thr.start()
    return wrapper




import os
import requests

def get_matched_bdating_wallet_info(consumer_wallet: str):
    endpoint = os.environ['BDATING_WALLET_API_ENDPOINT']
    endpoint_key = os.environ['BDATING_WALLET_API_KEY']
    return requests.get(
        url=f"{endpoint}match",
        params={
            'user_wallet': consumer_wallet,
        },
        headers={
            "x-api-key": endpoint_key
        }
    ).json()



import os
import json
import logging
import boto3
#from bdating_common.json_logger import setup_logger


def collect_info_for_apps(
    namespace: str = None,
    whoami: str = None,
    aws_profile: str = None,
    aws_region: str = None,
    event_bus_topic: str = None,
):
    log = logging.getLogger(__name__) 
    if namespace is None:
        log.debug('Finding whoami from environments NAMESPACE')
    namespace = os.environ.get('NAMESPACE', 'local')
    if whoami is None:
        whoami = os.environ.get('WHO_AM_I', 'unknown')
        if not whoami.startswith(namespace):
            whoami = f"{namespace}-{whoami}"
    if aws_region is None:
        log.debug(
            'Trying to find aws region information from environment AWS_REGION')
        aws_region = os.environ.get('AWS_REGION')
    if aws_region is None:
        log.debug('Using default aws region: sydney (ap-southeast-2)')
        aws_region = 'ap-southeast-2'
    if aws_profile is None:
        log.debug(
          'Trying to find aws aws profile information from environment AWS_PROFILE')
        aws_profile = os.environ.get('AWS_PROFILE')
    
    session = boto3.Session(profile_name=aws_profile, region_name=aws_region)
    if event_bus_topic is None:
        log.debug(
          'Trying to find event bus from environment BUSINESS_EVENT_BUS_TOPIC')
        event_bus_topic = os.environ.get('BUSINESS_EVENT_BUS_TOPIC')
        if not event_bus_topic:
            log.debug('Trying to construct event bus from account')
            sts_client = session.client('sts')
            client_info = sts_client.get_caller_identity()
            account = client_info.get('Account')
            event_bus_topic = f"arn:aws:sns:{aws_region}:{account}:{namespace}-EventBus"
    result = dict(
        namespace=namespace,
        aws_profile=aws_profile,
        aws_region=aws_region,
        event_bus_topic=event_bus_topic,
        session=session,
        whoami=whoami
    )
    log.info('All settings: ' + json.dumps(result, default=str))
    return result



import setuptools
import bdating_common
with open("README.md") as fp:
    long_description = fp.read()

setuptools.setup(
    name="bdating-common",
    version=bdating_common.__version__,

    description="Bdating common libraries standard library",
    long_description=long_description,
    long_description_content_type="text/markdown",

    author="Zac",
    author_email="zac@bdating.io",
    package_dir={"": "."},
    packages=setuptools.find_packages(),
    package_data={'': ['data/*', '*.cer', '*.yml']},
    url='https://github.com/bdating-io/bdating-python-common',
    install_requires=[
        "boto3<2",
        "click-log",
        "click",
        "dnspython==2.2.0",
        "PyMySQL",
        "elasticsearch>8",
        "fastapi",
        "PyJWT==2.1.0",
        "cryptography == 3.4.7",
        "redis<5",
        "auth0-python",
        "pyyaml",
        "web3<6,>5.0",
        "requests",
        "wsproto==1.2.0",
        "aioredis<2"
    ],

    python_requires=">=3.6",

    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
    entry_points={
        'console_scripts': [
            'report-metrics=bdating_common.report_metrics:report',
            'report-events=bdating_common.report_events:report'
        ],
    }
)



#!/usr/bin/env python
# coding=utf-8
import logging
import json
import time
import os


class JSONFormatter(logging.Formatter):
    """ JSON formatter """
    def format(self, record):
        """ Format event info to json."""
        string_formatted_time = time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(record.created))
        obj = {}
        obj["message"] = record.msg
        obj["level"] = record.levelname
        obj["@timestamp"] = f"{string_formatted_time}.{record.msecs:3.0f}Z"
        obj["@version"] = "1"
        obj["epoch_time"] = record.created
        if hasattr(record, "custom_logging"):
            for key, value in record.custom_logging.items():
                obj[key] = value
        return json.dumps(obj)


def setup_logger(name: str) -> object:
    """ Create logging object."""
    if name:
        logger = logging.getLogger(name)
    else:
        logger = logging.getLogger(__name__)

    handler = logging.StreamHandler()
    formatter = JSONFormatter()
    handler.setFormatter(formatter)

    logger.addHandler(handler)
    logger.propagate=False

    if "DEBUG" in os.environ and os.environ["DEBUG"] == "true":
        logger.setLevel(logging.DEBUG)  # pragma: no cover
    else:
        logger.setLevel(os.getenv("LOG_LEVEL", logging.INFO))
        logging.getLogger("boto3").setLevel(logging.WARNING)
        logging.getLogger("botocore").setLevel(logging.WARNING)

    return logger